import io
from asyncio import Lock

import nio
from slidge import BaseSession
from slidge.db import GatewayUser
from slidge.util.types import PseudoPresenceShow
from slixmpp.exceptions import XMPPError
from slixmpp.types import ResourceDict

from .contact import Contact, Roster
from .group import MUC, Bookmarks, Participant
from .matrix import Client

Sender = Contact | Participant
Recipient = MUC | Contact


class Session(BaseSession[Contact]):
    bookmarks: Bookmarks
    contacts: Roster

    MESSAGE_IDS_ARE_THREAD_IDS = True

    def __init__(self, user: GatewayUser) -> None:
        super().__init__(user)
        self.events_to_ignore = set[str]()
        self.send_lock = Lock()
        f = self.user.legacy_module_data
        self.matrix = Client(f["homeserver"], f["user_id"], self)  # type:ignore[arg-type]

    async def login(self) -> str:
        await self.matrix.login_token()
        await self.matrix.listen()
        self.contacts.user_legacy_id = self.matrix.user_id
        return f"Logged in as {self.matrix.user}"

    async def logout(self) -> None:
        self.matrix.stop_listen()

    async def on_presence(
        self,
        resource: str,
        show: PseudoPresenceShow,
        status: str,
        resources: dict[str, ResourceDict],
        merged_resource: ResourceDict | None,
    ) -> None:
        if not merged_resource:
            resp = await self.matrix.set_presence("offline")
        else:
            resp = await self.matrix.set_presence(
                PRESENCE_DICT[merged_resource["show"]],
                merged_resource["status"] or None,
            )
        self.log.debug("Set presence response: %s", resp)

    async def on_avatar(
        self,
        bytes_: bytes | None,
        hash_: str | None,
        type_: str | None,
        width: int | None,
        height: int | None,
    ) -> None:
        if bytes_:
            resp, _ = await self.matrix.upload(
                io.BytesIO(bytes_), type_, filesize=len(bytes_)
            )
            self.log.debug("Upload response: %s %r", type(resp), resp)
            if not isinstance(resp, nio.UploadResponse):
                raise XMPPError("internal-server-error", str(resp))
            uri = resp.content_uri
        else:
            uri = ""
        await self.matrix.set_avatar(uri)


PRESENCE_DICT: dict[PseudoPresenceShow, str] = {
    "": "online",
    "chat": "online",
    "away": "unavailable",
    "xa": "unavailable",
    "dnd": "unavailable",
}



import json
from typing import TypedDict, cast

import nio
from nio.crypto import InboundGroupSession, OlmDevice, TrustState
from slidge.command import Command, CommandAccess, FormField, TableResult
from slidge.command.base import (
    FormRecipient,
    FormSession,
    FormValues,
    MUCCommand,
    Option,
)
from slidge.command.categories import CONTACTS, GROUPS
from slidge.util.types import LegacyAttachment
from slixmpp import JID
from slixmpp.exceptions import XMPPError

from .group import MUC
from .session import Session


class ListSpaces(Command[Session]):
    NAME = "🌌 Matrix spaces"
    CHAT_COMMAND = NODE = "spaces"
    CATEGORY = GROUPS
    HELP = "List the matrix spaces you're part of"
    ACCESS = CommandAccess.USER_LOGGED

    async def run(
        self,
        session: Session | None,
        _ifrom: JID | None,
        *args: str,
    ) -> FormSession[Session]:
        assert session is not None

        spaces = list[nio.MatrixRoom]()
        for room in session.matrix.rooms.values():
            if room.children:
                spaces.append(room)
        spaces = sorted(spaces, key=lambda r: r.name)
        return FormSession(
            title=self.NAME,
            instructions="Choose a space to list its children rooms. "
            "NB: as of now, you can also see rooms that you are a member of.",
            handler=self.finish,
            handler_args=(spaces,),
            fields=[
                FormField(
                    "space",
                    label="Matrix space",
                    type="list-single",
                    options=[
                        {"label": room.display_name, "value": str(i)}
                        for i, room in enumerate(spaces)
                    ],
                )
            ],
        )

    @staticmethod
    async def finish(
        form_values: FormValues,
        session: Session,
        _ifrom: JID | None,
        rooms: list[nio.MatrixRoom],
    ) -> TableResult:
        space = rooms[int(form_values["space"])]  # type:ignore[arg-type]
        mucs = list[MUC]()
        for room_id in space.children:
            try:
                mucs.append(await session.bookmarks.by_legacy_id(room_id))
            except XMPPError:
                continue

        mucs = sorted(mucs, key=lambda muc: muc.name or "Unnamed room")
        return TableResult(
            fields=[FormField("name"), FormField("jid", type="jid-single")],
            description=f"Rooms of '{space.display_name}'",
            jids_are_mucs=True,
            items=[
                {"name": muc.name or "Unnamed room", "jid": str(muc.jid)}
                for muc in mucs
            ],
        )


class ManageTrust(Command[Session]):
    NAME = "🤝 Manage trust in devices"
    CATEGORY = CONTACTS
    CHAT_COMMAND = NODE = "verify"
    HELP = "Manage which OLM keys you trust or not."
    ACCESS = CommandAccess.USER_LOGGED

    HUMAN_STATES = {0: "unset", 1: "verified", 2: "blacklisted", 3: "ignored"}

    def __human_device(self, d: OlmDevice, state: bool = True) -> str:
        r = f"{d.ed25519} of {d.user_id}"
        if state:
            return r + f" ({self.HUMAN_STATES[d.trust_state.value]})"
        return r

    async def run(
        self,
        session: Session | None,
        _ifrom: JID | None,
        *args: str,
    ) -> FormSession[Session] | str:
        assert session is not None

        devices = list[OlmDevice](session.matrix.olm.device_store)
        device_dict = {d.id: d for d in devices}

        # this part if for chat commands only
        if args:
            if args[0] == "all":
                return await self.step2(
                    {
                        "device": list(device_dict.keys()),
                        "new_state": "verified",
                    },
                    session,
                    None,
                    device_dict,
                )

            else:
                return await self.step2(
                    {"device": args[0].upper(), "new_state": "verified"},
                    session,
                    None,
                    device_dict,
                )

        return FormSession(
            title=self.NAME,
            instructions="Choose the session(s) which trust state you want to change",
            handler=self.step2,
            handler_args=(device_dict,),
            fields=[
                FormField(
                    "device",
                    label="Device(s)",
                    type="list-multi",
                    options=[
                        {"label": self.__human_device(d), "value": d.id}
                        for d in devices
                    ],
                ),
                FormField(
                    "new_state",
                    label="What new status do you want to give the selected devices?",
                    type="list-single",
                    options=[
                        {"label": v, "value": v} for v in self.HUMAN_STATES.values()
                    ],
                ),
            ],
        )

    async def step2(
        self,
        form_values: FormValues,
        session: Session,
        _ifrom: JID | None,
        devices: dict[str, OlmDevice],
    ) -> str:
        new_state = form_values["new_state"]
        matrix = session.matrix
        result = ""
        for device_name in form_values["device"]:  # type:ignore
            assert isinstance(device_name, str)
            device = devices[device_name]
            change = False
            if new_state == "unset":
                if device.trust_state == TrustState.verified:
                    change = matrix.unverify_device(device)
                elif device.trust_state == TrustState.ignored:
                    change = matrix.unignore_device(device)
                elif device.trust_state == TrustState.blacklisted:
                    change = matrix.unblacklist_device(device)
            elif new_state == "verified":
                change = session.matrix.verify_device(device)
            elif new_state == "blacklisted":
                change = session.matrix.blacklist_device(device)
            elif new_state == "ignored":
                change = session.matrix.ignore_device(device)

            if change:
                result += (
                    f"\nThe status of {self.__human_device(device, False)} "
                    f"is now {new_state}."
                )
            else:
                result += (
                    f"\nThe status of {self.__human_device(device, False)} "
                    f"has not changed."
                )

        return result or "Nothing was changed."


class ChangePassword(Command[Session]):
    NAME = "🔑 Change Password"
    CHAT_COMMAND = NODE = "change-password"
    HELP = "Change your Matrix account password."
    ACCESS = CommandAccess.USER_LOGGED

    async def run(
        self,
        session: Session | None,
        _ifrom: JID | None,
        *args: str,
    ) -> FormSession[Session] | str:
        assert session is not None

        return FormSession(
            title=self.NAME,
            instructions="Enter your current and new Matrix credentials.",
            handler=self.change_password,
            fields=[
                FormField(
                    "current_password",
                    label="Current Password",
                    type="text-private",
                    required=True,
                ),
                FormField(
                    "new_password",
                    label="New Password",
                    type="text-private",
                    required=True,
                ),
                FormField(
                    "confirm_new_password",
                    label="Confirm New Password",
                    type="text-private",
                    required=True,
                ),
            ],
        )

    async def change_password(
        self,
        form_values: FormValues,
        session: Session,
        _ifrom: JID | None,
    ) -> str:
        current_password = form_values["current_password"]
        new_password = form_values["new_password"]
        confirm_new_password = form_values["confirm_new_password"]

        if new_password != confirm_new_password:
            raise XMPPError("not-acceptable", "New passwords do not match.")

        user_id = session.matrix.user_id
        homeserver = session.matrix.homeserver
        username = user_id.split(":")[0][1:]

        url = f"{homeserver.rstrip('/')}/_matrix/client/r0/account/password"
        payload = {
            "auth": {
                "type": "m.login.password",
                "identifier": {"type": "m.id.user", "user": username},
                "password": current_password,
            },
            "new_password": new_password,
        }
        headers = {
            "Authorization": f"Bearer {session.matrix.access_token}",
            "Content-Type": "application/json",
        }

        async with session.http.post(
            url,
            json=payload,
            headers=headers,
        ) as response:
            if response.status == 200:
                return "Password changed successfully!"

            error_data = await response.json()
            error_msg = error_data.get("error", "Unknown error")
            errcode = error_data.get("errcode", "M_UNKNOWN")
            raise XMPPError(
                "internal-server-error",
                f"Server error: [{errcode}] {error_msg}",
            )


class DeleteDevices(Command[Session]):
    NAME = "🔑 Delete devices"
    CHAT_COMMAND = NODE = "delete-devices"
    HELP = "Remove devices from your Matrix account."
    ACCESS = CommandAccess.USER_LOGGED

    def _device_label(self, device: nio.Device) -> str:
        return f"{device.id}: {device.display_name or '(no name)'}"

    async def run(
        self,
        session: Session | None,
        _ifrom: JID | None,
        *args: str,
    ) -> FormSession[Session] | str:
        assert session is not None

        response = await session.matrix.devices()

        if isinstance(response, nio.DevicesError):
            raise XMPPError(
                "internal-server-error",
                f"Failed to list devices: {response.message}",
            )

        if not (devices := response.devices):
            return "No devices found."

        device_opts: list[Option] = [
            {"label": self._device_label(d), "value": d.id} for d in devices
        ]
        return FormSession(
            title=self.NAME,
            instructions="Select device to delete from your account.",
            handler=self.delete_step,
            fields=[
                FormField(
                    "devices",
                    label="Devices to delete",
                    type="list-multi",
                    options=device_opts,
                    required=True,
                ),
            ],
        )

    async def delete_step(
        self,
        form_values: FormValues,
        session: Session,
        _ifrom: JID | None,
    ) -> str | FormSession[Session]:
        devices_to_delete: list[str] = form_values["devices"]  # type:ignore

        response = await session.matrix.delete_devices(devices_to_delete)

        if isinstance(response, nio.DeleteDevicesError):
            raise XMPPError(
                "internal-server-error",
                f"Failed to delete devices: {response.message}",
            )

        if isinstance(response, nio.DeleteDevicesAuthResponse):
            return FormSession(
                title=self.NAME,
                instructions="Authentication required to delete devices.",
                handler=self.delete_with_auth,
                handler_args=(devices_to_delete, response),
                fields=[
                    FormField(
                        "password",
                        label="Password",
                        type="text-private",
                        required=True,
                    ),
                ],
            )

        # XXX: this branch untested; my server requires auth.
        deleted_names = ", ".join(devices_to_delete)
        return f"Successfully deleted devices: {deleted_names}"

    async def delete_with_auth(
        self,
        form_values: FormValues,
        session: Session,
        _ifrom: JID | None,
        devices_to_delete: list[str],
        auth_response: nio.DeleteDevicesAuthResponse,
    ) -> str:
        """Prompts for a password and completes the deletion.

        nio doesn't store the password, so we need to prompt for it.
        """
        password = form_values["password"]
        user_id = session.matrix.user_id
        username = user_id.split(":")[0][1:]

        auth = {
            "type": "m.login.password",
            "identifier": {"type": "m.id.user", "user": username},
            "password": password,
            "session": auth_response.session,
        }

        response = await session.matrix.delete_devices(devices_to_delete, auth)

        if isinstance(response, nio.DeleteDevicesError):
            raise XMPPError(
                "internal-server-error",
                f"Failed to delete devices: {response.message}",
            )

        names = ", ".join(devices_to_delete)
        return f"Successfully deleted devices: {names}"


class JoinRoom(Command[Session]):
    NAME = "🚪 Join Matrix room"
    CHAT_COMMAND = NODE = "join-room"
    CATEGORY = GROUPS
    HELP = "Join a Matrix room by name or alias."
    ACCESS = CommandAccess.USER_LOGGED

    async def run(
        self,
        session: Session | None,
        _ifrom: JID | None,
        *args: str,
    ) -> FormSession[Session] | str:
        assert session is not None

        if args:  # For chat command.
            room = args[0]
            return await self.join_room({"room": room}, session, None)

        return FormSession(
            title=self.NAME,
            instructions="Enter the Matrix room name (format: #public:example.com)",
            handler=self.join_room,
            fields=[
                FormField(
                    "room",
                    label="Room name.",
                    type="text-single",
                    required=True,
                ),
            ],
        )

    async def join_room(
        self,
        form_values: FormValues,
        session: Session,
        _ifrom: JID | None,
    ) -> str:
        room = form_values["room"]

        response = await session.matrix.join(room)

        if isinstance(response, nio.JoinError):
            raise XMPPError(
                "internal-server-error",
                f"Failed to join room: {response.message}",
            )

        room_id = response.room_id

        muc = await session.bookmarks.by_legacy_id(room_id)

        try:
            await muc.add_to_bookmarks()
        except XMPPError as e:
            # Reconnecting to the gateway should make the room show up.
            return f"Joined room {room}, but couldn't add bookmark: {e}"

        return f"Successfully joined room {room} (ID: {room_id})"


class SessionKeyData(TypedDict):
    algorithm: str
    sender_key: str
    sender_claimed_keys: dict[str, str]
    forwarding_curve25519_key_chain: list[str]
    room_id: str
    session_id: str
    session_key: str


class ExportRoomKeys(Command[Session]):
    NAME = "🔐 Export room keys"
    CHAT_COMMAND = NODE = "export-keys"
    HELP = "Export all Megolm session keys."
    ACCESS = CommandAccess.USER_LOGGED

    async def run(
        self,
        session: Session | None,
        _ifrom: JID | None,
        *args: str,
    ) -> str:
        assert session is not None

        matrix = session.matrix
        sessions_list: list[SessionKeyData] = []

        inbound_store = matrix.store.load_inbound_group_sessions()
        for sess in inbound_store:
            sessions_list.append(
                {
                    "algorithm": "m.megolm.v1.aes-sha2",
                    "sender_key": sess.sender_key,
                    "sender_claimed_keys": {"ed25519": sess.ed25519},
                    "forwarding_curve25519_key_chain": sess.forwarding_chain,
                    "room_id": sess.room_id,
                    "session_id": sess.id,
                    "session_key": sess.export_session(sess.first_known_index),
                }
            )

        device_curve25519 = matrix.olm.account.identity_keys["curve25519"]
        device_ed25519 = matrix.olm.account.identity_keys["ed25519"]
        for room_id, sess in matrix.olm.outbound_group_sessions.items():
            sessions_list.append(
                {
                    "algorithm": "m.megolm.v1.aes-sha2",
                    "sender_key": device_curve25519,
                    "sender_claimed_keys": {"ed25519": device_ed25519},
                    "forwarding_curve25519_key_chain": [],
                    "room_id": room_id,
                    "session_id": sess.session_id,
                    "session_key": sess.export_session(sess.first_known_index),
                }
            )

        data = json.dumps(sessions_list, indent=2).encode("utf-8")

        attachment = LegacyAttachment(
            data=data,
            name="matrix_room_keys.json",
            content_type="application/json",
        )
        await session.xmpp.send_file(attachment, mto=session.user_jid)

        return f"Room keys exported ({len(sessions_list)} sessions)"


class ImportRoomKeys(Command[Session]):
    NAME = "🔐 Import room keys"
    CHAT_COMMAND = NODE = "import-keys"
    HELP = "Import Megolm session keys."
    ACCESS = CommandAccess.USER_LOGGED

    async def run(
        self,
        session: Session | None,
        _ifrom: JID | None,
        *args: str,
    ) -> FormSession[Session] | str:
        assert session is not None

        return FormSession(
            title=self.NAME,
            instructions="Paste the JSON content from the exported room keys file.",
            handler=self.import_keys,
            fields=[
                FormField(
                    "json_content",
                    label="Room keys JSON",
                    type="text-multi",
                    required=True,
                ),
            ],
        )

    async def import_keys(
        self,
        form_values: FormValues,
        session: Session,
        _ifrom: JID | None,
    ) -> str:
        raw_lines = cast(list[str], form_values["json_content"])
        json_content = "".join(raw_lines)

        try:
            sessions_list = json.loads(json_content)
        except json.JSONDecodeError as e:
            raise XMPPError(
                "bad-request",
                f"Invalid JSON: {e}",
            )

        if not isinstance(sessions_list, list):
            raise XMPPError(
                "bad-request",
                "Invalid format: expected a JSON array of sessions",
            )

        matrix = session.matrix

        imported = 0
        skipped = 0
        failed = 0

        store = matrix.olm.inbound_group_store
        for key_data in sessions_list:
            try:
                room_id = key_data["room_id"]
                session_id = key_data["session_id"]
                sender_key = key_data["sender_key"]

                # Nio always returns True, even for existing keys.
                # See: https://github.com/matrix-nio/matrix-nio/issues/561
                existing = store.get(room_id, sender_key, session_id)
                if existing:
                    skipped += 1
                    session.log.debug(
                        f"Existed: {room_id} sender={sender_key} session={session_id}"
                    )
                    continue

                inbound_session = InboundGroupSession.import_session(
                    session_key=key_data["session_key"],
                    signing_key=key_data["sender_claimed_keys"]["ed25519"],
                    sender_key=sender_key,
                    room_id=room_id,
                    forwarding_chain=key_data.get(
                        "forwarding_curve25519_key_chain", []
                    ),
                )

                if store.add(inbound_session):
                    matrix.store.save_inbound_group_session(inbound_session)
                    imported += 1
                    session.log.debug(
                        f"Imported: {room_id} "
                        f"sender={sender_key} session={session_id} "
                        f"first_known_index={inbound_session.first_known_index}"
                    )
                else:
                    skipped += 1
                    session.log.debug(
                        f"Skipped: {room_id} sender={sender_key} session={session_id}"
                    )

            except KeyError as e:
                session.log.warning(f"Missing required field {e} in session entry")
                failed += 1
            except Exception as e:
                session.log.warning(f"Failed to import session: {e}")
                failed += 1

        session.log.debug(f"Imported {imported}, skipped {skipped}, failed {failed}")

        # Previous backfills discarded messages which required these decryption keys.
        # Re-fetch them for processing with imported keys.

        # Slidge's deduplication (by stanza_id/legacy_id) ensures that
        # already-backfilled messages are updated rather than duplicated.
        imported_rooms = {key_data["room_id"] for key_data in sessions_list}
        rebackfilled = 0
        for room_id in imported_rooms:
            try:
                muc = await session.bookmarks.by_legacy_id(room_id)
            except XMPPError:
                continue  # No longer present in room.
            muc.stored.history_filled = False
            try:
                await muc.backfill()
                muc.commit()
                rebackfilled += 1
            except Exception as e:
                session.log.warning("Failed to re-backfill %s: %s", room_id, e)
            finally:
                muc.stored.history_filled = True
            # FIXME: change to info once we've finished testing.
            session.log.error(f"Re-backfilled {room_id}.")

        return (
            f"Imported {imported} new session keys,"
            f" {skipped} already present, {failed} failed."
            f" Re-backfilled {rebackfilled} rooms."
        )


class TrustUnverified(MUCCommand[MUC]):
    NAME = "🔐 Trust unverified devices"
    NODE = CHAT_COMMAND = "trust-unverified"
    HELP = "Toggle whether to trust unverified devices in this room"

    @staticmethod
    async def run(muc: MUC, *args: str) -> FormRecipient[MUC]:
        current = muc.ignore_unverified_devices
        current_str = "ENABLED" if current else "DISABLED"
        return FormRecipient(
            title=TrustUnverified.NAME,
            instructions=(
                f"Room: {muc.name or muc.legacy_id}\n\n"
                f"Current setting: {current_str}\n\n"
                "When enabled, messages will be sent to unverified devices "
                "in this room. This is less secure but may be necessary if "
                "room members have unverified devices.\n\n"
                "When disabled, only verified devices will "
                "receive your messages."
            ),
            handler=TrustUnverified.toggle,
            fields=[
                FormField(
                    "ignore_unverified",
                    label="Trust unverified devices in this room?",
                    type="boolean",
                    value=str(current).lower(),
                    required=True,
                )
            ],
        )

    @staticmethod
    async def toggle(muc: MUC, form_values: FormValues) -> str:
        new_value = form_values["ignore_unverified"] == "true"
        muc.ignore_unverified_devices = new_value
        if new_value:
            return (
                f"Room '{muc.name or muc.legacy_id}': Unverified devices are now TRUSTED.\n"
                "Your messages will be sent to all devices, even if unverified.\n"
                "This reduces end-to-end encryption security."
            )
        else:
            return (
                f"Room '{muc.name or muc.legacy_id}': Unverified devices are now IGNORED.\n"
                "Your messages will only be sent to verified devices.\n"
                "This is the recommended secure setting."
            )



import io
import typing
from collections.abc import AsyncIterator

import nio
from slidge.db.meta import JSONSerializable
from slidge.group import LegacyBookmarks, LegacyMUC, LegacyParticipant, MucType
from slidge.util.types import (
    ChatState,
    HoleBound,
    MucAffiliation,
    XMPPAttachmentMessage,
    XMPPMessage,
)
from slixmpp import Presence
from slixmpp.exceptions import XMPPError

from matridge.contact import Contact
from matridge.matrix import Client

from . import config
from .util import MatrixMixin, get_content, server_timestamp_to_datetime

if typing.TYPE_CHECKING:
    from .session import Session


class Participant(MatrixMixin, LegacyParticipant[Contact]):
    session: "Session"
    muc: "MUC"
    contact: Contact

    def set_affiliation_from_power_level(self, power_level: int) -> None:
        if power_level == 100:
            self.affiliation = "owner"
            self.role = "moderator"
        elif power_level >= 50:
            self.affiliation = "admin"
            self.role = "moderator"
        else:
            self.affiliation = "member"
            self.role = "participant"

    async def on_kick(self, reason: str | None) -> None:
        """Kick a user from the Matrix room."""
        user_id = self.contact.legacy_id
        resp = await self.session.matrix.room_kick(
            self.contact.legacy_id, user_id, reason
        )
        if isinstance(resp, nio.RoomKickError):
            raise XMPPError("internal-server-error", str(resp))

    async def on_set_affiliation(
        self,
        affiliation: MucAffiliation,
        reason: str | None,
        nickname: str | None,
    ) -> None:
        """Handle XMPP affiliation changes and translate to Matrix."""
        user_id = self.contact.legacy_id
        room_id = self.muc.legacy_id

        if affiliation == "member":
            resp = await self.session.matrix.room_invite(room_id, user_id)
            if isinstance(resp, nio.RoomInviteError):
                raise XMPPError("internal-server-error", str(resp))
            await self.muc._set_power_level(user_id, 0)
        elif affiliation == "outcast":
            resp = await self.session.matrix.room_ban(room_id, user_id, reason)
            if isinstance(resp, nio.RoomBanError):
                raise XMPPError("internal-server-error", str(resp))
        elif affiliation == "none":
            # Remove privileges first to avoid leaving user unbanned with power.
            await self.muc._set_power_level(user_id, 0)

            if await self.muc._is_banned(user_id):
                resp = await self.session.matrix.room_unban(room_id, user_id)
                if isinstance(resp, nio.RoomUnbanError):
                    raise XMPPError("internal-server-error", str(resp))
            else:
                resp = await self.session.matrix.room_kick(room_id, user_id, reason)
                if isinstance(resp, nio.RoomKickError):
                    raise XMPPError("internal-server-error", str(resp))
        elif affiliation == "owner":
            await self.muc._set_power_level(user_id, 100)
        elif affiliation == "admin":
            await self.muc._set_power_level(user_id, 50)
        else:
            self.log.warning("Affiliation %s not implemented.", affiliation)


class MUC(LegacyMUC[Participant]):
    session: "Session"
    _ignore_unverified_devices: bool | None = None

    @property
    def matrix(self) -> Client:
        return self.session.matrix

    def serialize_extra_attributes(self) -> JSONSerializable | None:
        """Save custom room settings to database."""
        if self._ignore_unverified_devices is None:
            return None
        return {"ignore_unverified_devices": self._ignore_unverified_devices}

    def deserialize_extra_attributes(self, data: JSONSerializable) -> None:
        """Restore custom room settings from database."""
        if ignore_unverified_devices := data.get("ignore_unverified_devices"):
            assert isinstance(ignore_unverified_devices, bool | None)
            self._ignore_unverified_devices = ignore_unverified_devices

    @property
    def ignore_unverified_devices(self) -> bool:
        """Get effective setting: room-specific or global fallback."""
        if self._ignore_unverified_devices is not None:
            return self._ignore_unverified_devices
        # Fallback to global setting
        return bool(self.session.user.preferences.get("trust_everything", False))

    @ignore_unverified_devices.setter
    def ignore_unverified_devices(self, value: bool) -> None:
        """Set room-specific setting and persist to database."""
        self._ignore_unverified_devices = value
        self.commit()

    async def get_room(self) -> nio.MatrixRoom:
        if room := self.session.matrix.rooms.get(self.legacy_id):
            return room
        # FIXME: invited_rooms is not persisted across restarts.
        if room := self.session.matrix.invited_rooms.get(self.legacy_id):
            return room
        raise XMPPError("item-not-found", f"No room named {self.legacy_id}")

    async def get_message(self, msg_id: str) -> nio.Event | None:
        return await self.session.matrix.get_event(self.legacy_id, msg_id)

    async def update_info(self) -> None:
        room = await self.get_room()

        if new := room.replacement_room:
            localpart = await self.session.bookmarks.legacy_id_to_jid_username(new)
            raise XMPPError(
                "redirect", f"xmpp:{localpart}@{self.xmpp.boundjid.bare}?join"
            )
        self.log.debug("Children: %s", room.children)
        if room.children:
            raise XMPPError("bad-request", "This is not a real room but a 'space'")

        self.user_nick = room.user_name(self.session.matrix.user_id)

        self.log.debug("User nick: %s", self.user_nick)
        self.name = room.display_name
        self.log.debug("Avatar: %s", room.room_avatar_url)
        self.n_participants = room.member_count
        self.subject = room.topic

        if room.join_rule == "public":
            self.type = MucType.CHANNEL_NON_ANONYMOUS
        else:
            self.type = MucType.GROUP

        if room.room_avatar_url:
            self.avatar = await self.session.matrix.mxc_to_http(room.room_avatar_url)

        elif room.member_count == 2:
            # if 1:1 set room avatar to other users's avatar
            for user_id, user in list(room.users.items()):
                if user_id == self.session.matrix.user_id:
                    continue
                if user.avatar_url is None:
                    break
                self.avatar = await self.session.matrix.mxc_to_http(user.avatar_url)
                break

    async def fill_participants(self) -> AsyncIterator[Participant]:
        room = await self.get_room()
        if not room.members_synced:
            resp = await self.session.matrix.joined_members(self.legacy_id)
            if isinstance(resp, nio.JoinedMembersResponse):
                self.log.debug(
                    "Joined members response: %s participants", len(resp.members)
                )
            else:
                self.log.debug("Joined members error: %s", resp)

        power_levels = room.power_levels.users
        i = 0

        for user_id, user in list(room.users.items()):
            power_level = power_levels.get(user_id, 0)
            if power_level < 50 and i > config.MAX_PARTICIPANTS_FETCH:
                continue
            try:
                p = await self.get_participant_by_legacy_id(user.user_id)
            except XMPPError:
                continue
            p.set_affiliation_from_power_level(power_level)
            if power_level < 50:
                i += 1
            yield p

    async def backfill(
        self,
        after: HoleBound | None = None,
        before: HoleBound | None = None,
    ) -> None:
        after_date = None if after is None else after.timestamp
        before_date = None if before is None else before.timestamp
        self.log.info("Backfilling from {%s} until {%s}.", after_date, before_date)

        total = 0
        undecryptable = 0
        skipped = 0
        filled = 0
        reactions = 0

        async for event in self.session.matrix.fetch_history(self.legacy_id):
            total += 1
            if not isinstance(event, (nio.RoomMessage | nio.ReactionEvent)):
                self.log.debug("Not back-filling with %s", type(event))
                if isinstance(event, nio.MegolmEvent):
                    undecryptable += 1
                continue

            when = server_timestamp_to_datetime(event)
            if after_date is not None and when <= after_date:
                # Events are yielded in reverse chronological order.
                # If this one is too old, then all subsequent ones are too.
                break
            if before_date is not None and when >= before_date:
                continue

            try:
                participant = await self.session.matrix.get_participant(
                    await self.get_room(), event
                )
            except XMPPError as e:
                # Maybe deleted profiles?
                self.log.warning(
                    "Failed to fetch participant %s, using matrix ID as nickname: %s",
                    event.sender,
                    e,
                )
                participant = await self.get_participant(event.sender)

            if isinstance(event, nio.RoomMessage):
                await participant.send_matrix_message(event, archive_only=True)
                filled += 1
            elif isinstance(event, nio.ReactionEvent):
                msg_id = await self.session.matrix.get_original_id(
                    self.legacy_id,
                    event.reacts_to,
                )
                participant.react(msg_id, [event.key], archive_only=True)
                reactions += 1

            if filled >= config.MAX_HISTORY_FETCH:
                break

        self.log.info(
            "Backfilled: %d events, %d filled, %d reactions, %d undecryptable, %d unhandled.",
            total,
            filled,
            reactions,
            undecryptable,
            skipped,
        )

    async def join(self, join_presence: Presence) -> None:
        """Called when user joins the room. Join on Matrix side if invited."""
        response = await self.session.matrix.join(self.legacy_id)
        if isinstance(response, nio.JoinResponse):
            self.log.debug(f"Successfully joined Matrix room {self.legacy_id}")
        else:
            self.log.warning(f"Failed to join Matrix room {self.legacy_id}: {response}")
        await super().join(join_presence)

    async def _is_banned(self, user_id: str) -> bool:
        resp = await self.session.matrix.room_get_state_event(
            self.legacy_id,
            "m.room.member",
            user_id,
        )
        if isinstance(resp, nio.RoomGetStateEventError):
            # M_NOT_FOUND means user has no membership state in this room.
            if resp.status_code == "M_NOT_FOUND":
                return False
            raise XMPPError("internal-server-error", str(resp))

        # If we don't compare types first, mypy things this returns Any.
        membership = resp.content.get("membership")
        return isinstance(membership, str) and membership == "ban"

    async def _set_power_level(self, user_id: str, power_level: int) -> None:
        """Set a user's power level in the Matrix room."""
        # This seems overwrites any concurrent changes (even to other users) submitted
        # by another operator. This is a limitation in Matrix's API.

        resp = await self.session.matrix.room_get_state_event(
            self.legacy_id, "m.room.power_levels", ""
        )
        if isinstance(resp, nio.RoomGetStateEventError):
            raise XMPPError("internal-server-error", str(resp))

        content = resp.content
        users = content.setdefault("users", {})
        users[user_id] = power_level

        resp = await self.session.matrix.room_put_state(
            self.legacy_id, "m.room.power_levels", content
        )
        if isinstance(resp, nio.RoomPutStateError):
            raise XMPPError("internal-server-error", str(resp))

    async def on_message(self, message: XMPPMessage) -> str | None:
        if message.attachments:
            return await self.__on_attachment(message)

        if message.replace:
            content = {
                "msgtype": "m.text",
                "body": "* " + message.body,
                "m.new_content": get_content(message.body, message.mentions),
                "m.relates_to": {
                    "rel_type": "m.replace",
                    "event_id": message.replace,
                },
            }
        else:
            content = get_content(message.body, message.mentions)
        await self.__relates_to(content, message)
        return await self.__room_send(content)

    async def __relates_to(
        self,
        content: dict[str, typing.Any],
        message: XMPPMessage,
    ) -> None:
        relates_to = dict[str, typing.Any]()
        if message.reply:
            relates_to["m.in_reply_to"] = {
                "event_id": await self.matrix.get_original_id(
                    self.legacy_id, message.reply.msg_id
                )
            }
        if message.thread:
            relates_to["rel_type"] = "m.thread"
            relates_to["event_id"] = message.thread
        if relates_to:
            content["m.relates_to"] = relates_to

    async def __room_send(
        self,
        content: dict[str, typing.Any],
        message_type: str = "m.room.message",
    ) -> str:
        await self.matrix.room_typing(self.legacy_id, False)
        async with self.session.send_lock:
            response = await self.matrix.room_send(
                self.legacy_id,
                message_type=message_type,
                content=content,
                ignore_unverified_devices=self.ignore_unverified_devices,
            )
            return await self.__handle_response(response)

    async def __handle_response(self, response: nio.Response) -> str:
        self.log.debug("Send response: %s", response)
        if isinstance(response, nio.RoomSendError):
            raise XMPPError("internal-server-error", str(response))
        assert isinstance(response, nio.RoomSendResponse)
        self.session.events_to_ignore.add(response.event_id)
        assert isinstance(response.event_id, str)
        return response.event_id

    async def __on_attachment(self, message: XMPPAttachmentMessage) -> str:
        if message.replace:
            raise XMPPError(
                "feature-not-implemented",
                "Replacing a message with a new message containing attachments is not supported",
            )

        # TODO: handle multiple attachments in here, when:
        # - an XMPP client actually implements that;
        # - and slidge core supports that too.
        url = message.attachments[0].url
        filename = url.split("/")[-1]
        async with message.attachments[0].get() as http_response:
            content_type = http_response.content_type
            resp, _ = await self.matrix.upload(
                io.BytesIO(await http_response.read()),
                content_type,
                filename,
                filesize=http_response.content_length,
            )
        self.log.debug("Upload response: %s %r", type(resp), resp)
        if not isinstance(resp, nio.UploadResponse):
            raise XMPPError("internal-server-error", str(resp))
        content = {
            "msgtype": "m.image" if content_type.startswith("image") else "m.file",
            "body": message.body or filename,
            "url": resp.content_uri,
        }
        await self.__relates_to(content, message)
        return await self.__room_send(content)

    async def on_chat_state(self, chat_state: ChatState, thread: str | None) -> None:
        match chat_state:
            case "composing":
                await self.matrix.room_typing(self.legacy_id)
            case "paused":
                await self.matrix.room_typing(self.legacy_id, False)

    async def on_displayed(self, legacy_msg_id: str, thread: str | None) -> None:
        resp = await self.matrix.update_receipt_marker(self.legacy_id, legacy_msg_id)
        self.log.debug("Displayed response: %s", resp)

    async def on_react(
        self, legacy_msg_id: str, emojis: list[str], thread: str | None
    ) -> None:
        new_emojis = set(emojis)
        with self.matrix.reactions.session() as session:
            old_emojis = await self.matrix.reactions.get(
                session,
                self.legacy_id,
                legacy_msg_id,
                self.matrix.user_id,
                with_event_ids=True,
            )
            for old_emoji, event in old_emojis.items():
                if old_emoji in new_emojis:
                    new_emojis.remove(old_emoji)
                else:
                    await self.on_retract(event, thread=None)
                    self.matrix.reactions.remove(session, self.legacy_id, event)

            for emoji in new_emojis:
                content = {
                    "m.relates_to": {
                        "rel_type": "m.annotation",
                        "event_id": legacy_msg_id,
                        "key": emoji,
                    },
                }

                i: str = await self.__room_send(content, "m.reaction")
                await self.matrix.reactions.add(
                    session,
                    self.legacy_id,
                    legacy_msg_id,
                    self.matrix.user_id,
                    emoji,
                    i,
                )
            session.commit()

    async def on_retract(self, legacy_msg_id: str, thread: str | None) -> None:
        resp = await self.matrix.room_redact(self.legacy_id, legacy_msg_id)
        self.log.debug("Redact response: %s", resp)
        if isinstance(resp, nio.RoomRedactError):
            raise XMPPError("internal-server-error", str(resp))
        assert isinstance(resp, nio.RoomRedactResponse)
        self.session.events_to_ignore.add(resp.event_id)

    async def on_leave(self) -> None:
        resp = await self.matrix.room_leave(self.legacy_id)
        self.log.debug("Leave response: %s", resp)
        # This prevents the room being re-listed when the user calls the
        # "groups" command which in turn calls Bookmarks.fill().
        del self.matrix.rooms[self.legacy_id]


class Bookmarks(LegacyBookmarks[MUC]):
    session: "Session"

    async def fill(self) -> None:
        self.log.debug("Filling rooms")
        self.log.debug(f"Joined rooms: {self.session.matrix.rooms.keys()}")
        self.log.debug(f"Invited rooms: {self.session.matrix.invited_rooms.keys()}")
        for room in self.session.matrix.rooms:
            try:
                muc = await self.by_legacy_id(room)
                await muc.add_to_bookmarks()
                self.log.debug(f"Successfully processed room {room}: {muc}")
            except XMPPError as e:
                self.log.debug(
                    "%s is not a group chat or trouble getting it: %r", room, e
                )

        # Relay Matrix invitations as XMPP room invites.
        for room_id in self.session.matrix.invited_rooms:
            try:
                muc = await self.by_legacy_id(room_id)
                self.log.debug(f"Relaying Matrix invitation to XMPP for {room_id}")
                self.session.send_gateway_invite(muc)
            except XMPPError as e:
                self.log.debug("Failed to relay invitation for %s: %r", room_id, e)
